package org.intranet.games.minesweeper;
// Copyright 2002 Chris Dailey
// License terms at http://www.gnu.org/licenses/lgpl.html

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/** 
 *
 * @author Chris Dailey
 */
public final class PlayingField
  implements Square.SquareListener, java.io.Serializable
{
  public PlayingField()
  {
    this(8, 8, 10);
  }

  public static final class Status
  {
    private String name;
    public static final Status initted = new Status("initted");
    public static final Status inprogress = new Status("in progress");
    public static final Status won = new Status("over(won)");
    public static final Status lost = new Status("over(lost)");
    Status(String s) { name = s; }
    public String toString() { return name; }
  }

  public void changeConfiguration(int w, int h, int numMines)
  {
    squares = new Square[w][];
    for (int i = 0; i < w; i++)
    {
      squares[i] = new Square[h];
      for (int j = 0; j < h; j++)
      {
        squares[i][j] = new Square(this, i, j);
        squares[i][j].addSquareListener(this);
      }
    }
    for (int i = 0; i < w; i++)
    {
      for (int j = 0; j < h; j++)
      {
        Square s = squares[i][j];
        for (int x = -1; x <= 1; x++)
        {
          for (int y = -1; y <= 1; y++)
          {
            int newX = x + i;
            int newY = y + j;
            if ((newX >= 0) && (newX < w) &&
                (newY >= 0) && (newY < h))
            {
              s.addNeighbor(squares[newX][newY]);
            }
          }
        }
      }
    }
    mines = new Mine[numMines];
    for (int i = 0; i < numMines; i++)
      mines[i] = new Mine();
    notifyConfigurationChanged();
    init();
  }

  public void init()
  {
    timer.init();
    squaresLeft = getWidth() * getHeight();
    // initting the squares should come after updating squaresLeft, else
    // you'll get undesired behavior from the callbacks of the events fired
    // from the initialization of the squares
    for (int i = getWidth() - 1; i >= 0; i--)
    {
      for (int j = getHeight() - 1; j >= 0; j--)
        squares[i][j].init();
    }
    flags.clear();
    status = Status.initted;
    notifyStatusChange();
    notifyCounterUpdated();
  }

  public int getWidth() { return squares.length; }
  public int getHeight() { return squares[0].length; }
  public Square getSquare(int x, int y) { return squares[x][y]; }
  public int getNumMines() { return mines.length; }
  public Status getStatus() { return status; }
  public boolean isOver() { return status == Status.won || status == Status.lost; }

  public void squareStateChanged(Square.SquareEvent se)
  {
    Square s = (Square)se.getSource();
    if (s.getState() == Square.State.uncovered)
    {
      if (status == Status.initted)
      {
        status = Status.inprogress;
        placeMines();
        timer.start();
        notifyStatusChange();
      }
      squaresLeft--;
      if ((squaresLeft - mines.length) == 0)
      {
        status = Status.won;
        endGame();
        notifyStatusChange();
      }
    }
    else if (s.getState() == Square.State.threatened)
    {
      squaresLeft++;
    }
    notifySquareChanged(s);
  }

  public void squareMineUncovered(Square.SquareEvent se)
  {
      status = Status.lost;
      endGame();
      notifyStatusChange();
  }

  public class PlayingFieldEvent
    extends java.util.EventObject
  {
    public PlayingFieldEvent()
    {
      super(PlayingField.this);
    }
  }

  public interface PlayingFieldListener
  {
    public void playingFieldSquareChanged(Square s);
    public void playingFieldConfigurationChanged(PlayingFieldEvent pfe);
    public void playingFieldCounterUpdated(int value);
    public void playingFieldStatusChanged(Status s);
  }

  public synchronized void addPlayingFieldListener(PlayingFieldListener l)
  {
    if (!playingFieldListeners.contains(l))
      playingFieldListeners.add(l);
  }

  public synchronized void removePlayingFieldListener(PlayingFieldListener l)
  {
    playingFieldListeners.remove(l);
  }

  public synchronized void addTimerListener(Timer.TimerListener l)
  {
    timer.addTimerListener(l);
  }

  public synchronized void removeTimerListener(Timer.TimerListener l)
  {
    timer.removeTimerListener(l);
  }

  // PROTECTED INTERFACE
  protected void notifyConfigurationChanged()
  {
    Iterator i = playingFieldListeners.iterator();
    while (i.hasNext())
    {
      PlayingFieldListener pfl = (PlayingFieldListener)i.next();
      pfl.playingFieldConfigurationChanged(playingFieldEvent);
    }
  }

  protected void notifyStatusChange()
  {
    Iterator i = playingFieldListeners.iterator();
    while (i.hasNext())
    {
      PlayingFieldListener pfl = (PlayingFieldListener)i.next();
      pfl.playingFieldStatusChanged(status);
    }
  }

  protected void notifyCounterUpdated()
  {
    int counter = mines.length - flags.size();
    Iterator i = playingFieldListeners.iterator();
    while (i.hasNext())
    {
      PlayingFieldListener pfl = (PlayingFieldListener)i.next();
      pfl.playingFieldCounterUpdated(counter);
    }
  }

  protected void notifySquareChanged(Square s)
  {
    Iterator i = playingFieldListeners.iterator();
    while (i.hasNext())
    {
      PlayingFieldListener pfl = (PlayingFieldListener)i.next();
      pfl.playingFieldSquareChanged(s);
    }
  }

  // PACKAGE INTERFACE
  final PlayingFieldEvent playingFieldEvent = new PlayingFieldEvent();

  Flag obtainFlag()
  {
    Flag f = new Flag();
    flags.add(f);
    notifyCounterUpdated();
    return f;
  }

  void returnFlag(Flag f)
  {
    flags.remove(f);
    notifyCounterUpdated();
  }

  PlayingField(int w, int h, int numMines)
  {
    r.setSeed(System.currentTimeMillis());
    changeConfiguration(w, h, numMines);
    init();
  }

  // PRIVATE INTERFACE
  private static final java.util.Random r = new java.util.Random();
  private Timer timer = new Timer();
  private Mine[] mines;
  private List flags = new ArrayList();
  /**
  * [width][height], or if you prefer, [x][y]
  */
  private Square[][] squares;
  private int squaresLeft;
  private List playingFieldListeners = new ArrayList();
  private Status status = Status.initted;

  private void placeMines()
  {
    for (int i = 0; i < mines.length; i++)
    {
      boolean found = false;
      while (!found)
      {
        int x = r.nextInt(squares.length);
        int y = r.nextInt(squares[0].length);
        Square s = squares[x][y];
        if (!s.hasMine() && s.getState() != Square.State.uncovered)
        {
          found = true;
          s.setMine(mines[i]);
        }
      }
    }
  }

  private void endGame()
  {
    timer.stop();
    // find all squares and tell them to end
    for (int x = 0; x < getWidth(); x++)
    {
      for (int y = 0; y < getHeight(); y++)
      {
        Square sq = squares[x][y];
        sq.end();
      }
    }
  }
}