package org.intranet.games.minesweeper;
// Copyright 2002 Chris Dailey
// License terms at http://www.gnu.org/licenses/lgpl.html

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/** 
* The square that lives on the playingfield.
* @author Chris Dailey
*/
public final class Square
{
  // PUBLIC INTERFACE
  public static abstract class State
  {
    private String name;

    void mark(Square s) { }
    void uncover(Square s) { }
    void threaten(Square s) { }
    void unthreaten(Square s) { }
    void scare(Square s) { }
    void unscare(Square s) { }

    void end(Square s) { s.setState(coveredEnd); }
    void init(Square s)
    {
      s.mFlag = null;
      s.mMine = null;
      s.mMineNeighbors = -1;
      s.setState(covered);
    }

    private State(String s) { name = s; }
    public String toString() { return name; }
    public static final State flagged = new State("flagged")
    {
      final void mark(Square s) { s.putFlag(); s.setState(question); }
      final void end(Square s) { s.setState(coveredFlagEnd); }
    };
    public static final State coveredEnd = new State("coveredEnd")
    {
      final void end(Square s) { }
    };
    public static final State coveredFlagEnd = new State("coveredFlagEnd")
    {
      final void end(Square s) { }
    };
    public static final State uncoveredEnd = new State("uncoveredEnd")
    {
      final void end(Square s) { }
    };
    public static final State covered = new State("covered")
    {
      final void mark(Square s)
      {
        s.obtainFlag();
        s.setState(flagged);
      }
      final void uncover(Square s)
      {
        s.setState(uncovered);
        if (s.hasMine())
          s.notifyMineUncovered();
        else if (s.getNumber() == 0)
        {
          Iterator neighborIterator = s.getNeighbors();
          while (neighborIterator.hasNext())
          {
            Square neighbor = (Square)neighborIterator.next();
            neighbor.uncover();
          }
        }
      }
      final void scare(Square s) { s.setState(scared); }
    };
    public static final State question = new State("question")
    {
      final void mark(Square s) { s.setState(covered); }
    };
    public static final State uncovered = new State("uncovered")
    {
      final void threaten(Square s)
      {
        s.setState(threatened);
        Iterator neighbors = s.getNeighbors();
        while (neighbors.hasNext())
        {
          Square neighbor = (Square)neighbors.next();
          neighbor.scare();
        }
      }
      final void mark(Square s)
      {
        s.uncoverNeighbors();
      }
      void end(Square s) { s.setState(uncoveredEnd); }
    };
    public static final State threatened = new State("threatened")
    {
      final void unthreaten(Square s)
      {
        s.setState(uncovered);
        Iterator neighbors = s.getNeighbors();
        while (neighbors.hasNext())
        {
          Square neighbor = (Square)neighbors.next();
          neighbor.unscare();
        }
//        s.uncoverNeighbors();
      }
    };
    public static final State scared = new State("scared")
    {
      final void unscare(Square s) { s.setState(covered); }
    };
  }


  public State getState() { return mState; }

  public void mark()
  {
    mState.mark(this);
  }

  public void uncover()
  {
    mState.uncover(this);
  }

  public void threaten()
  {
    mState.threaten(this);
  }
  public void unthreaten()
  {
    mState.unthreaten(this);
  }

  public PlayingField getPlayingField()
  {
    return mPlayingField;
  }

  public int getNumber()
  {
    if (mMineNeighbors == -1)
    {
      mMineNeighbors = 0;
      Iterator i = mNeighbors.iterator();
      while (i.hasNext())
      {
        Square s = (Square)i.next();
        if (s.hasMine())
          mMineNeighbors++;
      }
    }
    return mMineNeighbors;
  }

  public boolean hasMine()
  {
    return (mMine != null);
  }

  public class SquareEvent
    extends java.util.EventObject
  {
    public SquareEvent()
    {
      super(Square.this);
    }
  }

  public interface SquareListener
  {
    public void squareStateChanged(SquareEvent se);
    public void squareMineUncovered(SquareEvent se);
  }

  public synchronized void addSquareListener(SquareListener l)
  {
    if (!mSquareListeners.contains(l))
      mSquareListeners.add(l);
  }

  public synchronized void removeSquareListener(SquareListener l)
  {
    mSquareListeners.remove(l);
  }

  // PROTECTED INTERFACE
  protected void notifyStateChange()
  {
    Iterator i = mSquareListeners.iterator();
    while (i.hasNext())
    {
      SquareListener sl = (SquareListener)i.next();
      sl.squareStateChanged(mSquareEvent);
    }
  }

  protected void notifyMineUncovered()
  {
    Iterator i = mSquareListeners.iterator();
    while (i.hasNext())
    {
      SquareListener sl = (SquareListener)i.next();
      sl.squareMineUncovered(mSquareEvent);
    }
  }

  // PACKAGE INTERFACE
  Iterator getNeighbors() { return mNeighbors.iterator(); }

  void obtainFlag()
  {
    mFlag = mPlayingField.obtainFlag();
  }

  void putFlag()
  {
    mPlayingField.returnFlag(mFlag);
    mFlag = null; 
  }

  void end()
  {
    mState.end(this);
  }

  void setMine(Mine m)
  {
    mMine = m;
  }

  Square(PlayingField pf, int x, int y)
  {
    mPlayingField = pf;
    mX = x;
    mY = y;
  }

  public int getX() { return mX; }
  public int getY() { return mY; }

  void init()
  {
    mState.init(this);
  }

  void addNeighbor(Square s)
  {
    mNeighbors.add(s);
  }

  void scare()
  {
    mState.scare(this);
  }

  void unscare()
  {
    mState.unscare(this);
  }

  void uncoverNeighbors()
  {
    if (mState != State.uncovered)
      return;
    Iterator i = mNeighbors.iterator();
    int numNeighborFlags = 0;
    while (i.hasNext())
    {
      Square s = (Square)i.next();
      if (s.getState() == State.flagged)
        numNeighborFlags++;
    }
    if (getNumber() != numNeighborFlags)
      return;
    i = mNeighbors.iterator();
    while (i.hasNext())
    {
      Square s = (Square)i.next();
      if (s.getState() == State.question)
        s.mark();
      if (s.getState() == State.covered)
        s.uncover();
    }
  }

  // PRIVATE INTERFACE
  private final SquareEvent mSquareEvent = new SquareEvent();
  private List mSquareListeners = new ArrayList();
  private State mState = State.covered;
  private Flag mFlag;
  private int mX;
  private int mY;
  private Mine mMine;
  private List mNeighbors = new ArrayList(9);
  private PlayingField mPlayingField;
  private int mMineNeighbors = -1;

  private void setState(State s)
  {
    if (mState == s || mState == null) return;
    mState = s;
    notifyStateChange();
  }
}