package org.intranet.games.minesweeper;
// Copyright 2002 Chris Dailey
// License terms at http://www.gnu.org/licenses/lgpl.html

/** 
 *
 * @author Chris Dailey
 */
public final class Timer
{
  public interface TimerListener
  {
    public void timerValueChanged(int value);
  }
  public void addTimerListener(TimerListener tl)
  {
    if (!timerListeners.contains(tl))
      timerListeners.add(tl);
  }
  public void removeTimerListener(TimerListener tl)
  {
    timerListeners.remove(tl);
  }
  private java.util.List timerListeners = new java.util.ArrayList();
  protected void notifyTimerListeners(int seconds)
  {
    for (int i = 0; i < timerListeners.size(); i++)
    {
      TimerListener tl = (TimerListener)timerListeners.get(i);
      tl.timerValueChanged(seconds);
    }
  }

  // States:
  // - Initted
  //    init:NOP
  //    start->Running
  //    stop:NOP
  // - Running
  //    init->Initted
  //    stop->Stopped
  //    start:NOP
  // - Stopped
  //    init->Initted
  //    start->Running
  //    stop:NOP

  long time = 0L;
  long lastTime = 0L;
  State state = State.initted;

  Thread t = new Thread()
  {
    public void run()
    {
      while (true)
      {
        synchronized(this) {
          if (state == State.initted || state == State.stopped)
            try {
              wait();
            } catch (InterruptedException ie) {
              ie.printStackTrace();
            }
          else if (state == State.running)
          {
            long currentTime = System.currentTimeMillis();
            long diff = currentTime - lastTime;
            long newTime = time + diff;
            if ((time / 1000) != ((newTime) / 1000))
            {
              diff = 0L;
              notifyTimerListeners((int)(newTime/1000));
            }
            time = newTime;
            lastTime = currentTime;
            long sleepTime = 1000 - (time % 1000);
            if (sleepTime == 0) sleepTime = 1000;
            try {
              wait(sleepTime);
            } catch (InterruptedException ie) {
              ie.printStackTrace();
            }
          }
          else if (state == State.exiting)
            break;
        }
      }
    }
  };

  private static class State
  {
    String name;
    State(String s) { name = s; }
    static final State initted = new State("initted");
    static final State running = new State("running");
    static final State stopped = new State("stopped");
    static final State exiting = new State("exiting");
    public final String toString() { return name; }
  }

  public Timer()
  {
    t.start();
  }

  public synchronized void init()
  {
    synchronized(t)
    {
      time = 0L;
      lastTime = 0;
      state = State.initted;
      notifyTimerListeners(0);
    }
  }

  public synchronized void start()
  {
    synchronized(t)
    {
      if (state != State.running)
      {
        lastTime = System.currentTimeMillis();
        state = State.running;
        t.notify();
        notifyTimerListeners((int)(time/1000));
      }
    }
  }

  public synchronized void stop()
  {
    synchronized(t)
    {
      if (state == State.running)
      {
        state = State.stopped;
        long currentTime = System.currentTimeMillis();
        time += currentTime - lastTime;
        t.notify();
      }
    }
  }

  public synchronized void exit()
  {
    state = State.exiting;
    timerListeners = null;
  }
}